#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/parport.h>
#include <linux/slab.h>

MODULE_AUTHOR("Markus Heinz");
MODULE_DESCRIPTION("Returns the device id");
MODULE_LICENSE("GPL");

MODULE_PARM(portnumber, "i");

#define OK 0
#define ERROR 1

#define BUFLEN 256

int init_module(void);
void cleanup_module(void);
int get_device_id(void);

int portnumber = 0; /* default to parport0 */
unsigned char device_id[BUFLEN]; /* will store the device id */

int init_module(void) { 
  if (get_device_id() == OK) {
    printk(KERN_INFO "Device id of parport%d: %s\n", portnumber, device_id);
  }

  return OK;
}

void cleanup_module(void) {

}

/* This function retrieves the device of the device connected to the
 * specified parport 
 */

int get_device_id(void) {
  int ret;
  unsigned char length[2];
  int idlen;

  struct pardevice *dev = parport_open(portnumber, "Device ID Probe", 
				       NULL, NULL, NULL, 0, NULL);
  if (dev == NULL) {
    printk(KERN_INFO "Failed to open parport%d\n", portnumber);
    return ERROR;
  }

  ret = parport_claim_or_block(dev);
  if (ret < 0) {
    printk(KERN_INFO "Failed to claim parport%d\n", portnumber);
    parport_close(dev);
    return ERROR;
  }

  ret = parport_negotiate(dev->port, IEEE1284_MODE_COMPAT); 
  if (ret != 0) {
    printk(KERN_INFO "Failed to negotiate compat mode for parport%d\n",
	   portnumber);
    parport_release(dev);
    parport_close(dev);
    return ERROR;
  }

  ret = parport_negotiate(dev->port, IEEE1284_MODE_NIBBLE | IEEE1284_DEVICEID);
  if (ret  != 0) {
    printk(KERN_INFO "Failed to retrieve device id of parport%d\n", 
	   portnumber);
    parport_release(dev);
    parport_close(dev);
    return ERROR;
  }

  ret = parport_read(dev->port, length, 2);
  if (ret != 2) {
    printk(KERN_INFO "Could not get length of device id of parport%d\n", 
	   portnumber);
    parport_negotiate(dev->port, IEEE1284_MODE_COMPAT);
    parport_release(dev);
    parport_close(dev);
    return ERROR;
  }
    
  idlen = (length[0] << 8) + length[1] - 2;
  idlen = idlen < 255 ? idlen : 255;
  
  ret = parport_read(dev->port, device_id, idlen);
  device_id[ret] = '\0';

  if (ret != idlen) {
    printk(KERN_INFO "Length of device id of parport%d does not fit\n",
	   portnumber);
  }
    
  parport_negotiate(dev->port, IEEE1284_MODE_COMPAT);
  parport_release(dev);
  parport_close(dev);

  return OK;
}

