#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/parport.h>
#include <linux/proc_fs.h>
#include <linux/slab.h>

MODULE_AUTHOR("Markus Heinz");
MODULE_DESCRIPTION("Returns the available amount of ink in the printer");
MODULE_LICENSE("GPL");

MODULE_PARM(portnumber, "i");

#define OK 0
#define ERROR -1

#define BUFLEN 256
#define NR_TAGS 15

int init_module(void);
void cleanup_module(void);
void register_proc(void);
void unregister_proc(void);
int proc_read_black(char *, char **, off_t, int, int);
int proc_read_color(char *, char **, off_t, int, int);
void do_the_work(void);
int get_device_id(void);
int parse_device_id(void);
int parse_device_id_new_hp(int);
int parse_device_id_old_hp(int);
int axtoi(char *);
int atoi(char *);

int portnumber = 0; /* default to parport0 */
unsigned char device_id[BUFLEN]; /* will store the device id */
int black = -1; /* will store the level of black ink */
int color = -1; /* will store the level of color ink */
struct proc_dir_entry *proc_dir; /* this points to the /proc/ink_level
				  * directory 
				  */
unsigned char tags[NR_TAGS][BUFLEN]; /* This will store the tags of the device
				      * id, one tag per string.
				      */

int init_module(void) { 
  register_proc(); 

  return OK;
}

void cleanup_module(void) {
  unregister_proc(); 
}

void register_proc(void) {
  proc_dir = proc_mkdir("ink_level", &proc_root);
  create_proc_read_entry("black", S_IFREG | S_IRUGO, proc_dir,
			 (void *) proc_read_black, NULL);
  create_proc_read_entry("color", S_IFREG | S_IRUGO, proc_dir,
			 (void *) proc_read_color, NULL);
}

void unregister_proc(void) {
  remove_proc_entry("black", proc_dir);
  remove_proc_entry("color", proc_dir);
  remove_proc_entry("ink_level", &proc_root);
}

int proc_read_black(char *buffer, char **buffer_location, off_t offset,
		    int buffer_length, int zero) {
  int len = 0;
  static char my_buffer[80];

  if (offset > 0) {
    return 0;
  }

  do_the_work();
  len = sprintf(my_buffer, "%d\n", black);
  
  *buffer_location = my_buffer;
  return len;
}

int proc_read_color(char *buffer, char **buffer_location, off_t offset,
		    int buffer_length, int zero) {
  int len = 0;
  static char my_buffer[80];

  if (offset > 0) {
    return 0;
  }

  do_the_work();
  len = sprintf(my_buffer, "%d\n", color);
  
  *buffer_location = my_buffer;
  return len;
}

void do_the_work(void) {
  if (get_device_id() == OK) {
    parse_device_id();   
  }
}

/* This function retrieves the device id of the device connected to the
 * specified parport 
 */

int get_device_id(void) {
  int ret;
  unsigned char length[2];
  int idlen;

  struct pardevice *dev = parport_open(portnumber, "Device ID Probe", 
				       NULL, NULL, NULL, 0, NULL);
  if (dev == NULL) {
    printk(KERN_INFO "Failed to open parport%d\n", portnumber);
    return ERROR;
  }

  ret = parport_claim_or_block(dev);
  if (ret < 0) {
    printk(KERN_INFO "Failed to claim parport%d\n", portnumber);
    parport_close(dev);
    return ERROR;
  }

  ret = parport_negotiate(dev->port, IEEE1284_MODE_COMPAT); 
  if (ret != 0) {
    printk(KERN_INFO "Failed to negotiate compat mode for parport%d\n",
	   portnumber);
    parport_release(dev);
    parport_close(dev);
    return ERROR;
  }

  ret = parport_negotiate(dev->port, IEEE1284_MODE_NIBBLE | IEEE1284_DEVICEID);
  if (ret  != 0) {
    printk(KERN_INFO "Failed to retrieve device id of parport%d\n", 
	   portnumber);
    parport_release(dev);
    parport_close(dev);
    return ERROR;
  }

  ret = parport_read(dev->port, length, 2);
  if (ret != 2) {
    printk(KERN_INFO "Could not get length of device id of parport%d\n", 
	   portnumber);
    parport_negotiate(dev->port, IEEE1284_MODE_COMPAT);
    parport_release(dev);
    parport_close(dev);
    return ERROR;
  }
    
  idlen = (length[0] << 8) + length[1] - 2;
  idlen = idlen < 255 ? idlen : 255;
  
  ret = parport_read(dev->port, device_id, idlen);
  device_id[ret] = '\0';

  if (ret != idlen) {
    printk(KERN_INFO "Length of device id of parport%d does not fit\n",
	   portnumber);
  }
    
  parport_negotiate(dev->port, IEEE1284_MODE_COMPAT);
  parport_release(dev);
  parport_close(dev);

  return OK;
}

/* This function determines the type of printer attached and calls
 * the appropiate parser function 
 */

int parse_device_id(void) {
  int i = 0;
  int j = 0;
  unsigned char *c;
  
  /* init tags */

  for (i = 0; i < NR_TAGS; i++) {
    for (j = 0; j < BUFLEN; j++) {
      tags[i][j] = '\0';
    }
  }

  /* Tokenize the device id */

  i = 0;
  j = 0;
  c = device_id;
  
  while ((*c != '\0') && (i < NR_TAGS)) {
    j = 0;
    while ((*c != ';') && (j < BUFLEN)) {
      tags[i][j] = *c;
      c++;
      j++;
    }
    c++; /* Skip the ';' */
    i++;
  }

#ifdef DEBUG
  /* print all tags */

  for (i = 0; i < NR_TAGS; i++) {
    printk(KERN_DEBUG "%d: %s\n", i, tags[i]);
  }
#endif

  /* Check if we deal with a printer */

  /* Find the "CLS:" tag */

  for (i = 0; i < NR_TAGS; i++) {
    c = tags[i];
    if ((c[0] == 'C') && (c[1] == 'L') && (c[2] == 'S') && (c[3] == ':')) {
      break;
    }
  }

  if (i < NR_TAGS) {

    /* Check if it is "PRINTER" */

    if ((tags[i][4] != 'P') || (tags[i][5] != 'R') || (tags[i][6] != 'I') || 
	(tags[i][7] != 'N') || (tags[i][8] != 'T') || (tags[i][9] != 'E') || 
	(tags[i][10] != 'R')) {
      
#ifdef DEBUG
      printk(KERN_DEBUG "No printer found\n");
#endif
      
      return ERROR;
    }
  } else { 
    /* Find the "CLASS:" tag */
    
    for (i = 0; i < NR_TAGS; i++) {
      c = tags[i];
      if ((c[0] == 'C') && (c[1] == 'L') && (c[2] == 'A') && (c[3] == 'S') && 
	  (c[4] == 'S') && (c[5] == ':')) {
	break;
      }
    }

    if (i < NR_TAGS) {
    
      /* Check if it is "PRINTER" */
      
      if ((tags[i][6] != 'P') || (tags[i][7] != 'R') || (tags[i][8] != 'I') || 
	  (tags[i][9] != 'N') || (tags[i][10] != 'T') || 
	  (tags[i][11] != 'E') || (tags[i][12] != 'R')) {
      
#ifdef DEBUG
	printk(KERN_DEBUG "No printer found\n");
#endif
	
	return ERROR;
      }
    } else {

#ifdef DEBUG 
      printk(KERN_DEBUG "No device class found\n");
#endif

      return ERROR;
    }
  }

  /* Check for a new HP printer */

  /* Find the "S:" tag */

  for (i = 0; i < NR_TAGS; i++) {
    c = tags[i];
    if ((c[0] == 'S') && (c[1] == ':')) {
      break;
    }
  }

#ifdef DEBUG
  if (i < NR_TAGS) {
    printk(KERN_DEBUG "The \"S:\" tag has number %d\n", i);
  } else {
    printk(KERN_DEBUG "No \"S:\" tag found\n");
  }
#endif

  if (i<NR_TAGS) {
    return parse_device_id_new_hp(i);
  }
  
  /* Check for an old HP printer */

  /* Find the "VSTATUS:" tag */

  for (i = 0; i < NR_TAGS; i++) {
    c = tags[i];
    if ((c[0] == 'V') && (c[1] == 'S') && (c[2] == 'T') && (c[3] == 'A') &&
	(c[4] == 'T') && (c[5] == 'U') && (c[6] == 'S') && (c[7] ==':')) {
      break;
    }
  }

#ifdef DEBUG
  if (i < NR_TAGS) {
    printk(KERN_DEBUG "The \"VSTATUS:\" tag has number %d\n", i);
  } else {
    printk(KERN_DEBUG "No \"VSTATUS:\" tag found\n");
  }
#endif

  if (i<NR_TAGS) {
    return parse_device_id_old_hp(i);
  }

  /* Insert code to check for other printers here */
  
  return ERROR; /* No matching printer was found */
}

/* This function parses the device id of a new HP printer
 * for example HP Deskjet 5550 
 */

int parse_device_id_new_hp(int n) {
  char *s = tags[n];
  int length = 0;
  unsigned char b[3]; /* level of black ink as hexadecimal string */
  unsigned char c[3]; /* level of color ink as hexadecimal string */


  /* Determine the length of the string */

  while (*s != '\0') {
    length++;
    s++;
  }

  s = tags[n];
  
  b[0] = s[length - 10];
  b[1] = s[length -9];
  b[2] = '\0';

  c[0] = s[length - 2];
  c[1] = s[length - 1];
  c[2] = '\0';

  black = axtoi(b);
  color = axtoi(c);

#ifdef DEBUG
  printk(KERN_DEBUG "Black: %d%%, Color: %d%%\n", black, color);
#endif

  return OK;
}

/* This function parses the device id of an old HP printer
 * for example HP Photosmart 1000 
 */

int parse_device_id_old_hp(int n) {
  char *s = tags[n];
  int length = 0;
  unsigned char b[4]; /* level of black ink as decimal string */
  unsigned char c[4]; /* level of color ink as decimal string */

  /* Determine the length of the string */

  while (*s != '\0') {
    length++;
    s++;
  }

  s = tags[n];
  
  /* Check if "KP" and "CP" tags are there */

  if ((s[length - 11] == 'K') && (s[length - 10] == 'P') &&
      (s[length - 5]  == 'C') && (s[length - 4] == 'P')) {
    b[0] = s[length - 9];
    b[1] = s[length - 8];
    b[2] = s[length - 7];
    b[3] = '\0';
    
    c[0] = s[length - 3];
    c[1] = s[length - 2];
    c[2] = s[length - 1];
    c[3] = '\0';

    black = atoi(b);
    color = atoi(c);

#ifdef DEBUG
    printk(KERN_DEBUG "Black; %d%%, Color %d%%\n", black, color);
#endif

    return OK;
  } else {

#ifdef DEBUG
    printk(KERN_DEBUG "No ink level found\n");
#endif
    return ERROR;
  }
}

/* This function converts a string containing a two digit hexadecimal number
 * to an int 
 */

int axtoi(char* t) {
  int r = 0;

  switch (t[0]) {
  case '0':
    r = 0;
    break;
  case '1':
    r = 16;
    break;
  case '2':
    r = 32;
    break;
  case '3':
    r = 48;
    break;
  case '4':
    r = 64;
    break;
  case '5':
    r = 80;
    break;
  case '6':
    r = 96;
    break;
  case '7':
    r = 112;
    break;
  case '8':
    r = 128;
    break;
  case '9':
    r = 144;
    break;
  case 'a':
  case 'A':
    r = 160;
    break;
  case 'b':
  case 'B':
    r = 176;
    break;
  case 'c':
  case 'C':
    r = 192;
    break;
  case 'd':
  case 'D':
    r = 208;
    break;
  case 'e':
  case 'E':
    r = 224;
    break;
  case 'f':
  case 'F':
    r = 240;
    break;
  }

  switch (t[1]) {
  case '0':
    r += 0;
    break;
  case '1':
    r += 1;
    break;
  case '2':
    r += 2;
    break;
  case '3':
    r += 3;
    break;
  case '4':
    r += 4;
    break;
  case '5':
    r += 5;
    break;
  case '6':
    r += 6;
    break;
  case '7':
    r += 7;
    break;
  case '8':
    r += 8;
    break;
  case '9':
    r += 9;
    break;
  case 'a':
  case 'A':
    r += 10;
    break;
  case 'b':
  case 'B':
    r += 11;
    break;
  case 'c':
  case 'C':
    r += 12;
    break;
  case 'd':
  case 'D':
    r += 13;
    break;
  case 'e':
  case 'E':
    r += 14;
    break;
  case 'f':
  case 'F':
    r += 15;
    break;
  }

  return r;
}

/* This function converts a string containig a three digit decimal number
 * to an int 
 */

int atoi(char* t) {
  int r = 0;

  switch (t[0]) {
  case '0':
    r = 0;
    break;
  case '1':
    r = 100;
    break;
  case '2':
    r = 200;
    break;
  case '3':
    r = 300;
    break;
  case '4':
    r = 400;
    break;
  case '5':
    r = 500;
    break;
  case '6':
    r = 600;
    break;
  case '7':
    r = 700;
    break;
  case '8':
    r = 800;
    break;
  case '9':
    r = 900;
    break;
  }
  
  switch (t[1]) {
  case '0':
    r += 0;
    break;
  case '1':
    r += 10;
    break;
  case '2':
    r += 20;
    break;
  case '3':
    r += 30;
    break;
  case '4':
    r += 40;
    break;
  case '5':
    r += 50;
    break;
  case '6':
    r += 60;
    break;
  case '7':
    r += 70;
    break;
  case '8':
    r += 80;
    break;
  case '9':
    r += 90;
    break;
  }

  switch (t[2]) {
  case '0':
    r += 0;
    break;
  case '1':
    r += 1;
    break;
  case '2':
    r += 2;
    break;
  case '3':
    r += 3;
    break;
  case '4':
    r += 4;
    break;
  case '5':
    r += 5;
    break;
  case '6':
    r += 6;
    break;
  case '7':
    r += 7;
    break;
  case '8':
    r += 8;
    break;
  case '9':
    r += 9;
    break;
  }
  
  return r;
}
